unit UMap;

{$mode delphi}

interface

uses
  Classes, SysUtils, UGame, XMLRead, XMLWrite, DOM, Contnrs, UGeometry;

type
  TCellsDistance = class
    Cell1: TCell;
    Cell2: TCell;
    Distance: Double;
  end;

  { THexMap }

  THexMap = class(TMap)
  private
    function IsCellsPosNeighbor(CellPos1, CellPos2: TPoint): Boolean;
    procedure GetCellPosNeighbors(CellPos: TPoint; Neighbours: TCells);
    function GetHexagonPolygon(Pos: TPoint; Size: TPoint): TPointArray;
  public
    procedure LoadFromFile(FileName: string); override;
    procedure SaveToFile(FileName: string); override;
    function IsValidIndex(Index: TPoint): Boolean; override;
    procedure Generate; override;
  end;

  { TSquareMap }

  TSquareMap = class(TMap)
  private
    function GetSquarePolygon(Pos: TPoint; Size: TPoint): TPointArray;
  public
    function IsValidIndex(Index: TPoint): Boolean; override;
    procedure Generate; override;
  end;

  { TTriangleMap }

  TTriangleMap = class(TMap)
  private
    function GetTrianglePolygon(Pos: TPoint; Size: TPoint; Reverse: Boolean): TPointArray;
  public
    function IsValidIndex(Index: TPoint): Boolean; override;
    procedure Generate; override;
  end;

  { TVoronoiMap }

  TVoronoiMap = class(TMap)
  private
    function GetTrianglePolygon(Pos: TPoint; Size: TPoint; Reverse: Boolean): TPointArray;
  public
    procedure Generate; override;
  end;

implementation

{ THexMap }

function THexMap.GetHexagonPolygon(Pos: TPoint; Size: TPoint): TPointArray;
var
  Shift: TFloatPoint;
begin
  Shift := FloatPoint(0.5 * cos(30 / 180 * Pi), 0.5 * sin(30 / 180 * Pi));
  SetLength(Result, 6);
  Result[0] := Point(Trunc(Pos.X + 0 * Size.X), Trunc(Pos.Y - 0.5 * Size.Y));
  Result[1] := Point(Trunc(Pos.X + Shift.X * Size.X), Trunc(Pos.Y - Shift.Y * Size.Y));
  Result[2] := Point(Trunc(Pos.X + Shift.X * Size.X), Trunc(Pos.Y + Shift.Y * Size.Y));
  Result[3] := Point(Trunc(Pos.X + 0 * Size.X), Trunc(Pos.Y + 0.5 * Size.Y));
  Result[4] := Point(Trunc(Pos.X - Shift.X * Size.X), Trunc(Pos.Y + Shift.Y * Size.Y));
  Result[5] := Point(Trunc(Pos.X - Shift.X * Size.X), Trunc(Pos.Y - Shift.Y * Size.Y));
end;

function THexMap.IsCellsPosNeighbor(CellPos1, CellPos2: TPoint): Boolean;
var
  DX: Integer;
  DY: Integer;
  MinY: Integer;
begin
  if CellPos1.Y < CellPos2.Y then MinY:= CellPos1.Y
    else MinY := CellPos2.Y;
  DX := CellPos2.X - CellPos1.X;
  DY := CellPos2.Y - CellPos1.Y;
  Result := (Abs(DX) <= 1) and (Abs(DY) <= 1) and
  ((((MinY mod 2) = 1) and
    not ((DX = 1) and (DY = -1)) and
    not ((DX = -1) and (DY = 1))) or
    (((MinY mod 2) = 0) and
    not ((DX = -1) and (DY = -1)) and
    not ((DX = 1) and (DY = 1))));
  Result := Result and not ((CellPos1.X = CellPos2.X) and (CellPos1.Y = CellPos2.Y));
end;

procedure THexMap.LoadFromFile(FileName: string);
var
  Doc: TXMLDocument;
begin
  try
    ReadXMLFile(Doc, FileName);
    if Doc.DocumentElement.TagName <> 'Map' then
      raise Exception.Create('Invalid map format');
  finally
    Doc.Free;
  end;
  inherited LoadFromFile(FileName);
end;

procedure THexMap.SaveToFile(FileName: string);
var
  Doc: TXMLDocument;
  RootNode: TDOMNode;
begin
  try
    Doc := TXMLDocument.Create;
    RootNode := Doc.CreateElement('Map');
    Doc.Appendchild(RootNode);
    WriteXMLFile(Doc, FileName);
  finally
    Doc.Free;
  end;
  inherited SaveToFile(FileName);
end;

function THexMap.IsValidIndex(Index: TPoint): Boolean;
begin
  Result := (Index.X >= 0) and (Index.X < Size.X) and
    (Index.Y >= 0) and (Index.Y < Size.Y);
end;

procedure THexMap.GetCellPosNeighbors(CellPos: TPoint; Neighbours: TCells);
var
  X, Y: Integer;
begin
  Neighbours.Count := 0;
  for Y := -1 to 1 do
  for X := -1 to 1 do
  if IsValidIndex(Point(CellPos.X + X, CellPos.Y + Y)) and
  IsCellsPosNeighbor(CellPos, Point((CellPos.X + X), (CellPos.Y + Y))) then begin
    Neighbours.Add(TCell(Cells[(CellPos.Y + Y) * Size.X + (CellPos.X + X)]));
  end;
end;

procedure THexMap.Generate;
var
  X, Y: Integer;
  NewCell: TCell;
  PX, PY: Double;
begin
  Clear;

  // Allocate and init new
  Cells.Count := Size.Y * Size.X;
  for Y := 0 to Size.Y - 1 do
  for X := 0 to Size.X - 1 do begin
    NewCell := TCell.Create;
    NewCell.Map := Self;
    PX := X;
    PY := Y;
    if (Y and 1) = 1 then begin
      PX := PX + 0.5;
      //Y := Y + 0.5;
    end;
    NewCell.PosPx := Point(Trunc(PX * DefaultCellSize.X / HexCellMulX),
      Trunc(PY * DefaultCellSize.Y / HexCellMulY));
    NewCell.Polygon := GetHexagonPolygon(NewCell.PosPx, DefaultCellSize);
    NewCell.Id := GetNewCellId;
    Cells[Y * Size.X + X] := NewCell;
  end;

  // Generate neightbours
  for Y := 0 to Size.Y - 1 do
  for X := 0 to Size.X - 1 do
  with TCell(Cells[Y * Size.X + X]) do begin
    GetCellPosNeighbors(Point(X, Y), Neighbors);
  end;

  FPixelRect := CalculatePixelRect;
end;


{ TSquareMap }

procedure TSquareMap.Generate;
var
  X, Y: Integer;
  NewCell: TCell;
begin
  Clear;

  // Allocate and init new
  Cells.Count := Size.Y * Size.X;
  for Y := 0 to Size.Y - 1 do
  for X := 0 to Size.X - 1 do begin
    NewCell := TCell.Create;
    NewCell.Map := Self;
    NewCell.PosPx := Point(Trunc(X * DefaultCellSize.X * SquareCellMulX),
      Trunc(Y * DefaultCellSize.Y * SquareCellMulY));
    NewCell.Polygon := GetSquarePolygon(NewCell.PosPx, DefaultCellSize);
    NewCell.Id := GetNewCellId;
    Cells[Y * Size.X + X] := NewCell;
  end;

  // Generate neighbours
  for Y := 0 to Size.Y - 1 do
  for X := 0 to Size.X - 1 do
  with TCell(Cells[Y * Size.X + X]) do begin
    if IsValidIndex(Point(X + 1, Y + 0)) then
      Neighbors.Add(TCell(Cells[(Y + 0) * Size.X + (X + 1)]));
    if IsValidIndex(Point(X + 0, Y + 1)) then
      Neighbors.Add(TCell(Cells[(Y + 1) * Size.X + (X + 0)]));
    if IsValidIndex(Point(X - 1, Y + 0)) then
      Neighbors.Add(TCell(Cells[(Y + 0) * Size.X + (X - 1)]));
    if IsValidIndex(Point(X + 0, Y - 1)) then
      Neighbors.Add(TCell(Cells[(Y - 1) * Size.X + (X + 0)]));
  end;

  FPixelRect := CalculatePixelRect;
end;

function TSquareMap.IsValidIndex(Index: TPoint): Boolean;
begin
  Result := (Index.X >= 0) and (Index.X < Size.X) and
    (Index.Y >= 0) and (Index.Y < Size.Y);
end;

function TSquareMap.GetSquarePolygon(Pos: TPoint; Size: TPoint): TPointArray;
begin
  SetLength(Result, 4);
  Result[0] := Point(Trunc(Pos.X - Size.X / 2), Trunc(Pos.Y - Size.Y / 2));
  Result[1] := Point(Trunc(Pos.X + Size.X / 2), Trunc(Pos.Y - Size.Y / 2));
  Result[2] := Point(Trunc(Pos.X + Size.X / 2), Trunc(Pos.Y + Size.Y / 2));
  Result[3] := Point(Trunc(Pos.X - Size.X / 2), Trunc(Pos.Y + Size.Y / 2));
end;

{ TVoronoiMap }

function TVoronoiMap.GetTrianglePolygon(Pos: TPoint; Size: TPoint;
  Reverse: Boolean): TPointArray;
begin
  SetLength(Result, 0);
end;

function CompareDistance(C1, C2: Pointer): Integer;
begin
  if TCellsDistance(C1).Distance > TCellsDistance(C2).Distance then Result := 1
  else if TCellsDistance(C1).Distance < TCellsDistance(C2).Distance then Result := -1
  else Result := 0;
end;

function CompareDistanceReverse(C1, C2: Pointer): Integer;
begin
  if TCellsDistance(C1).Distance > TCellsDistance(C2).Distance then Result := -1
  else if TCellsDistance(C1).Distance < TCellsDistance(C2).Distance then Result := 1
  else Result := 0;
end;

procedure TVoronoiMap.Generate;
var
  X, Y: Integer;
  I1, I2: Integer;
  NewCell: TCell;
  CellsDistance: TObjectList; // TObjectList<TCellsDistance>
  NewCellDist: TCellsDistance;
  SelectedCells: TObjectList; // TObjectList<TCellsDistance>
  I, J: Integer;
  Intersected: Boolean;
  Intersection: TPoint;
  CellDistance: TCellsDistance;
  Cell: TCell;
  L1, L2: TLine;
  MP: TPoint;
  LinkLine: TLine;
  Polygon: TPolygon;
  //LeftLink: TCellsDistance;
  //CenterLink: TCellsDistance;
  //RightLink: TCellsDistance;
  LeftClosingLine1: TLine;
  LeftClosingLine2: TLine;
  RightClosingLine1: TLine;
  RightClosingLine2: TLine;
  CurrentAngle: Double;
  LeftAngle: Double;
  RightAngle: Double;
  CenterCell: TCell;
  LeftCell: TCell;
  RightCell: TCell;
  LeftIndex: Integer;
  RightIndex: Integer;
  LeftCenterCell: TCell;
  RightCenterCell: TCell;
  ChangesCount: Integer;
  LeftCellCommon: TCell;
  RightCellCommon: TCell;
  LeftText: string;
  RightText: string;
  NeighborCell: TCell;
begin
  Clear;
  RandSeed := 1234;

  // Allocate and init new cells
  Cells.Count := Size.Y * Size.X;
  for Y := 0 to Size.Y - 1 do
  for X := 0 to Size.X - 1 do begin
    NewCell := TCell.Create;
    NewCell.Map := Self;
    NewCell.PosPx := Point(Trunc(Random * Size.X * DefaultCellSize.X),
      Trunc(Random * Size.Y * DefaultCellSize.Y));
    SetLength(NewCell.Polygon, 1);
    NewCell.Polygon[0] := NewCell.PosPx;
    NewCell.Id := GetNewCellId;
    Cells[Y * Size.X + X] := NewCell;
  end;

  // Calculate distance between all cells
  CellsDistance := TObjectList.Create;
  for I1 := 1 to Cells.Count - 1 do
  for I2 := I1 + 1 to Cells.Count - 1 do begin
    NewCellDist := TCellsDistance.Create;
    NewCellDist.Cell1 := Cells[I1];
    NewCellDist.Cell2 := Cells[I2];
    NewCellDist.Distance := Distance(NewCellDist.Cell1.PosPx, NewCellDist.Cell2.PosPx);
    CellsDistance.Add(NewCellDist);
  end;
  CellsDistance.Sort(CompareDistanceReverse);

  // Keep shortest non-intersected cell pairs
  SelectedCells := TObjectList.Create;
  SelectedCells.OwnsObjects := False;
  I := CellsDistance.Count - 1;
  while I >= 0 do begin
    Intersected := False;
    for J := 0 to SelectedCells.Count - 1 do
    if (TCellsDistance(SelectedCells[J]).Cell1 <> TCellsDistance(CellsDistance[I]).Cell1)
    and (TCellsDistance(SelectedCells[J]).Cell2 <> TCellsDistance(CellsDistance[I]).Cell2)
    and (TCellsDistance(SelectedCells[J]).Cell1 <> TCellsDistance(CellsDistance[I]).Cell2)
    and (TCellsDistance(SelectedCells[J]).Cell2 <> TCellsDistance(CellsDistance[I]).Cell1) then begin
      L1 := TLine.Create(TCellsDistance(CellsDistance[I]).Cell1.PosPx,
        TCellsDistance(CellsDistance[I]).Cell2.PosPx);
      L2 := TLine.Create(TCellsDistance(SelectedCells[J]).Cell1.PosPx,
        TCellsDistance(SelectedCells[J]).Cell2.PosPx);
      if LineIntersect(L1, L2, Intersection) then begin
        if PointInRect(Intersection, L1.ToRect) and
        PointInRect(Intersection, L2.ToRect) then begin
          Intersected := True;
          Break;
        end;
      end;
    end;
    if not Intersected then SelectedCells.Add(CellsDistance[I]);
    Dec(I);
  end;

  // Add cell neighbors
  for I := 0 to SelectedCells.Count - 1 do begin
    CellDistance := TCellsDistance(SelectedCells[I]);
    CellDistance.Cell1.Neighbors.Add(CellDistance.Cell2);
    CellDistance.Cell2.Neighbors.Add(CellDistance.Cell1);
  end;

  FreeAndNil(SelectedCells);
  FreeAndNil(CellsDistance);
  SortNeighborsByAngle;

  while True do begin
  ChangesCount := 0;
  // Optimize link lines for lower angle
  for Cell in Cells do begin
    // Change link if lower angle can be achieved
    for I := 0 to Cell.Neighbors.Count - 1 do begin
      LeftCell := Cell.Neighbors[I];
      RightCell := Cell.Neighbors[(I + 1) mod Cell.Neighbors.Count];
      //LeftText := LeftCell.Neighbors.ToString;
      //RightText := RightCell.Neighbors.ToString;
      LeftIndex := LeftCell.Neighbors.IndexOf(Cell);
      RightIndex := RightCell.Neighbors.IndexOf(Cell);
      LeftCellCommon := LeftCell.Neighbors[(LeftIndex + LeftCell.Neighbors.Count - 1) mod LeftCell.Neighbors.Count];
      RightCellCommon := RightCell.Neighbors[(RightIndex + 1) mod RightCell.Neighbors.Count];
      if (LeftCellCommon = RightCell) and
      (RightCellCommon = LeftCell) then begin
        LeftCenterCell := LeftCell.Neighbors[(LeftIndex + LeftCell.Neighbors.Count - 2) mod LeftCell.Neighbors.Count];
        RightCenterCell := RightCell.Neighbors[(RightIndex + 2) mod RightCell.Neighbors.Count];

        if LeftCenterCell = RightCenterCell then begin
          CenterCell := LeftCenterCell;

          LeftClosingLine1 := TLine.Create(LeftCell.PosPx, Cell.PosPx);
          LeftClosingLine2 := TLine.Create(LeftCell.PosPx,
            CenterCell.PosPx);
          LeftAngle := SubAngle(LeftClosingLine1.GetAngle,
            LeftClosingLine2.GetAngle);

          RightClosingLine1 := TLine.Create(RightCell.PosPx, Cell.PosPx);
          RightClosingLine2 := TLine.Create(RightCell.PosPx,
            CenterCell.PosPx);
          RightAngle := SubAngle(RightClosingLine2.GetAngle,
            RightClosingLine1.GetAngle);

          LeftClosingLine1 := TLine.Create(Cell.PosPx, LeftCell.PosPx);
          RightClosingLine1 := TLine.Create(Cell.PosPx, RightCell.PosPx);
          CurrentAngle := SubAngle(RightClosingLine1.GetAngle,
            LeftClosingLine1.GetAngle);
          if (CurrentAngle > LeftAngle) and (CurrentAngle > RightAngle) then begin
            LeftCell.DisconnectFrom(RightCell);
            Cell.ConnectTo(CenterCell);
            SortNeighborsByAngle;
            Inc(ChangesCount);
          end;
        end;
        //end;
      end;
    end;
  end;
  if ChangesCount = 0 then Break;
  end;

  // Compute polygon around cells with sequence sorted by link angle
  for Cell in Cells do begin
    // Use whole map first for cell polygon
    if Cell.Neighbors.Count > 0 then begin
      Polygon := TPolygon.Create(Rect(0, 0,
        Size.X * DefaultCellSize.X, Size.Y * DefaultCellSize.Y));
      for NeighborCell in Cell.Neighbors do begin
        LinkLine := TLine.Create(Cell.PosPx, NeighborCell.PosPx);
        LinkLine.Distance := LinkLine.Distance - 4;
        MP := LinkLine.GetMiddle;
        // Create half plane vector
        L1 := TLine.Create(MP, Point(MP.X + LinkLine.GetSize.X,
          MP.Y + LinkLine.GetSize.Y));

        Polygon.CutLine(L1, Cell.PosPx);
      end;
      Cell.Polygon := Polygon.Points;
    end else SetLength(Cell.Polygon, 0);
  end;

  FPixelRect := CalculatePixelRect;
end;

{ TTriangleMap }

function TTriangleMap.GetTrianglePolygon(Pos: TPoint; Size: TPoint;
  Reverse: Boolean): TPointArray;
var
  Rev: Integer;
begin
  if Reverse then Rev := -1
    else Rev := 1;
  SetLength(Result, 3);
  Result[0] := Point(Trunc(Pos.X - Size.X / 2), Trunc(Pos.Y - (Size.Y * 0.8) / 2 * Rev));
  Result[1] := Point(Trunc(Pos.X + Size.X / 2), Trunc(Pos.Y - (Size.Y * 0.8) / 2 * Rev));
  Result[2] := Point(Trunc(Pos.X), Trunc(Pos.Y + (Size.Y * 1.2) / 2 * Rev));
end;

function TTriangleMap.IsValidIndex(Index: TPoint): Boolean;
begin
  Result := (Index.X >= 0) and (Index.X < Size.X) and
    (Index.Y >= 0) and (Index.Y < Size.Y);
end;

procedure TTriangleMap.Generate;
var
  X, Y: Integer;
  Rev: Integer;
  Reverse: Boolean;
  NewCell: TCell;
begin
  Clear;

  // Allocate and init new
  Cells.Count := Size.Y * Size.X;
  for Y := 0 to Size.Y - 1 do
  for X := 0 to Size.X - 1 do begin
    NewCell := TCell.Create;
    NewCell.Map := Self;
    Reverse := Boolean(X mod 2) xor Boolean(Y mod 2);
    if Reverse then Rev := -1
      else Rev := 1;
    NewCell.PosPx := Point(Trunc(X * DefaultCellSize.X * TriangleCellMulX),
      Trunc((Y * DefaultCellSize.Y * TriangleCellMulY) - (0.1 * Rev * DefaultCellSize.Y)));
    NewCell.Polygon := GetTrianglePolygon(NewCell.PosPx, DefaultCellSize, Reverse);
    NewCell.Id := GetNewCellId;
    Cells[Y * Size.X + X] := NewCell;
  end;

  // Generate neighbours
  for Y := 0 to Self.Size.Y - 1 do
  for X := 0 to Size.X - 1 do
  with TCell(Cells[Y * Size.X + X]) do begin
    if Boolean(X mod 2) xor Boolean(Y mod 2) then Rev := -1
      else Rev := 1;
    if IsValidIndex(Point(X + 1, Y + 0)) then
      Neighbors.Add(TCell(Cells[(Y + 0) * Size.X + (X + 1)]));
    if IsValidIndex(Point(X + 0, Y - 1 * Rev)) then
      Neighbors.Add(TCell(Cells[(Y - 1 * Rev) * Size.X + (X + 0)]));
    if IsValidIndex(Point(X - 1, Y + 0)) then
      Neighbors.Add(TCell(Cells[(Y + 0) * Size.X + (X - 1)]));
  end;

  FPixelRect := CalculatePixelRect;
end;


end.

